from __future__ import annotations

__all__ = ["BaseSequenceGenerator", "is_sequence_generator_config", "setup_sequence_generator"]

import logging
from abc import ABC, abstractmethod

from objectory import AbstractFactory
from objectory.utils import is_object_config
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


class BaseSequenceGenerator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to generate sequences.

    A child class has to implement the ``generate`` method.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandUniform
        >>> generator = RandUniform()
        >>> generator
        RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    @abstractmethod
    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        r"""Generates a batch of sequences.

        All the sequences in the batch must have the same length.

        Args:
        ----
            seq_len (int): Specifies the sequence length.
            batch_size (int, optional): Specifies the batch size.
                Default: ``1``
            rng (``torch.Generator`` or None, optional): Specifies
                an optional random number generator. Default: ``None``

        Returns:
        -------
            ``BatchedTensorSeq``: A batch of sequences. The data in the
                batch are represented by a ``torch.Tensor`` of shape
                ``(batch_size, sequence_length, *)`` where `*` means
                any number of dimensions.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.sequence import RandUniform
            >>> generator = RandUniform()
            >>> generator.generate(seq_len=12, batch_size=4)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """


def is_sequence_generator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseSequenceGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseSequenceGenerator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import is_sequence_generator_config
        >>> is_sequence_generator_config({"_target_": "startorch.sequence.RandUniform"})
        True
    """
    return is_object_config(config, BaseSequenceGenerator)


def setup_sequence_generator(generator: BaseSequenceGenerator | dict) -> BaseSequenceGenerator:
    r"""Sets up a sequence generator.

    The sequence generator is instantiated from its configuration by
    using the ``BaseSequenceGenerator`` factory function.

    Args:
    ----
        generator (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator or its configuration.

    Returns:
    -------
        ``BaseSequenceGenerator``: A sequence generator.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import setup_sequence_generator
        >>> setup_sequence_generator({"_target_": "startorch.sequence.RandUniform"})
        RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
    """
    if isinstance(generator, dict):
        logger.info(
            "Initializing a sequence generator from its configuration... "
            f"{str_target_object(generator)}"
        )
        generator = BaseSequenceGenerator.factory(**generator)
    if not isinstance(generator, BaseSequenceGenerator):
        logger.warning(f"generator is not a `BaseSequenceGenerator` (received: {type(generator)})")
    return generator



from __future__ import annotations

__all__ = ["TimeSeriesGenerator"]


from coola.utils import str_indent, str_mapping
from redcat import BatchDict
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.timeseries.base import BaseTimeSeriesGenerator


class TimeSeriesGenerator(BaseTimeSeriesGenerator):
    r"""Implements a generic time series generator.

    Args:
    ----
        sequences (dict): Specifies the sequence generators or their
            configurations.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandUniform
        >>> from startorch.timeseries import TimeSeries
        >>> generator = TimeSeries({"value": RandUniform(), "time": RandUniform()})
        >>> generator
        TimeSeriesGenerator(
          (value): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (time): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        BatchDict(
          (value): tensor([[...]], batch_dim=0, seq_dim=1)
          (time): tensor([[...]], batch_dim=0, seq_dim=1)
        )
    """

    def __init__(self, sequences: dict[str, BaseSequenceGenerator | dict]) -> None:
        super().__init__()
        self._sequences = {
            key: setup_sequence_generator(generator) for key, generator in sequences.items()
        }

    def __repr__(self) -> str:
        args = str_indent(str_mapping(self._sequences))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchDict:
        return BatchDict(
            {
                key: generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
                for key, generator in self._sequences.items()
            }
        )



__all__ = [
    "asinh_uniform",
    "cauchy",
    "exponential",
    "half_cauchy",
    "half_normal",
    "log_normal",
    "log_uniform",
    "normal",
    "rand_asinh_uniform",
    "rand_cauchy",
    "rand_exponential",
    "rand_half_cauchy",
    "rand_half_normal",
    "rand_log_normal",
    "rand_log_uniform",
    "rand_normal",
    "rand_poisson",
    "rand_trunc_cauchy",
    "rand_trunc_exponential",
    "rand_trunc_half_cauchy",
    "rand_trunc_half_normal",
    "rand_trunc_log_normal",
    "rand_trunc_normal",
    "rand_uniform",
    "trunc_cauchy",
    "trunc_exponential",
    "trunc_half_cauchy",
    "trunc_half_normal",
    "trunc_log_normal",
    "trunc_normal",
    "uniform",
]

from startorch.random.bounded import (
    asinh_uniform,
    log_uniform,
    rand_asinh_uniform,
    rand_log_uniform,
    rand_trunc_cauchy,
    rand_trunc_exponential,
    rand_trunc_half_cauchy,
    rand_trunc_half_normal,
    rand_trunc_log_normal,
    rand_trunc_normal,
    rand_uniform,
    trunc_cauchy,
    trunc_exponential,
    trunc_half_cauchy,
    trunc_half_normal,
    trunc_log_normal,
    trunc_normal,
    uniform,
)
from startorch.random.discrete import rand_poisson
from startorch.random.infinite import cauchy, normal, rand_cauchy, rand_normal
from startorch.random.semi_infinite import (
    exponential,
    half_cauchy,
    half_normal,
    log_normal,
    rand_exponential,
    rand_half_cauchy,
    rand_half_normal,
    rand_log_normal,
)



from __future__ import annotations

__all__ = ["PoissonSequenceGenerator", "RandPoissonSequenceGenerator"]


import torch
from coola.utils import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.random import rand_poisson
from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.sequence.constant import ConstantSequenceGenerator
from startorch.utils.conversion import to_tuple


class PoissonSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    Poisson distribution.

    The rates of the Poisson distribution are generated by the rate
    generator. The rate generator should return the rate for each value
    in the sequence. The rate values should be greater than 0.

    Args:
    ----
        rate (``BaseSequenceGenerator`` or dict): Specifies the
            rate generator or its configuration. The rate generator
            should return valid rate values.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandUniform, Poisson
        >>> generator = Poisson(rate=RandUniform(low=1.0, high=2.0))
        >>> generator
        PoissonSequenceGenerator(
          (rate): RandUniformSequenceGenerator(low=1.0, high=2.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(self, rate: BaseSequenceGenerator | dict) -> None:
        super().__init__()
        self._rate = setup_sequence_generator(rate)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"rate": self._rate}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            torch.poisson(
                self._rate.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                generator=rng,
            )
        )

    @classmethod
    def generate_uniform_rate(
        cls,
        min_rate: float = 0.01,
        max_rate: float = 1.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> PoissonSequenceGenerator:
        r"""Implements a sequence generator where the rates of the
        Poisson distribution are sampled from a uniform distribution.

        Args:
        ----
            min_rate (float, optional): Specifies the minimum rate
                value. Default: ``0.01``
            max_rate (float, optional): Specifies the maximum rate
                value. Default: ``1.0``
            feature_size (tuple or list or int, optional): Specifies the
                feature size. Default: ``1``

        Returns:
            ``Poisson``: A sequence generator
                where the rates of the Poisson distribution are
                sampled from a uniform distribution
                (``UniformConstantSequenceGenerator``).
        """
        # The import is here to do not generate circular dependencies
        from startorch.sequence.uniform import RandUniformSequenceGenerator

        return cls(
            rate=ConstantSequenceGenerator(
                generator=RandUniformSequenceGenerator(
                    low=min_rate,
                    high=max_rate,
                    feature_size=feature_size,
                )
            ),
        )


class RandPoissonSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    Poisson distribution.

    Args:
    ----
        rate (float, optional): Specifies the rate of the Poisson
            distribution. This value has to be greater than 0.
            Default: ``1.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``rate`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandPoisson
        >>> generator = RandPoisson(rate=1.0)
        >>> generator
        RandPoissonSequenceGenerator(rate=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        rate: float = 1.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        if rate <= 0:
            raise ValueError(f"rate has to be greater than 0 (received: {rate})")
        self._rate = float(rate)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(rate={self._rate}, feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_poisson(
                size=(batch_size, seq_len) + self._feature_size,
                rate=self._rate,
                generator=rng,
            )
        )



from __future__ import annotations

__all__ = ["RepeatPeriodicSequenceGenerator"]

import math

from coola.utils import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.periodic.sequence import BasePeriodicSequenceGenerator
from startorch.sequence import BaseSequenceGenerator, setup_sequence_generator


class RepeatPeriodicSequenceGenerator(BasePeriodicSequenceGenerator):
    r"""Implements a class to generate periodic sequences by using a
    ``BaseSequenceGenerator`` object and repeating the generated
    sequence.

    Args:
    ----
        sequence (``BaseSequenceGenerator`` or dict): Specifies
            a sequence generator or its configuration.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.periodic.sequence import Repeat
        >>> from startorch.sequence import RandUniform
        >>> generator = Repeat(RandUniform())
        >>> generator
        RepeatPeriodicSequenceGenerator(
          (generator): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, period=4, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(self, generator: BaseSequenceGenerator | dict) -> None:
        super().__init__()
        self._generator = setup_sequence_generator(generator)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"generator": self._generator}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, period: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return (
            self._generator.generate(seq_len=period, batch_size=batch_size, rng=rng)
            .repeat_along_seq(math.ceil(seq_len / period))
            .slice_along_seq(stop=seq_len)
        )



from __future__ import annotations

__all__ = ["get_random_seed", "get_torch_generator"]


import torch


def get_random_seed(seed: int) -> int:
    r"""Gets a random seed.

    Args:
    ----
        seed (int): Specifies a random seed to make the process
            reproducible.

    Returns:
    -------
        int: A random seed. The value is between ``-2 ** 63`` and
            ``2 ** 63 - 1``.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.seed import get_random_seed
        >>> get_random_seed(44)
        6176747449835261347
    """
    return torch.randint(
        -(2**63), 2**63 - 1, size=(1,), generator=get_torch_generator(seed)
    ).item()


def get_torch_generator(
    random_seed: int = 1, device: torch.device | str | None = "cpu"
) -> torch.Generator:
    r"""Creates a ``torch.Generator`` initialized with a given seed.

    Args:
    ----
        random_seed (int, optional): Specifies a random seed.
            Default: ``1``
        device (``torch.device`` or str or ``None``, optional):
            Specifies the desired device for the generator.
            Default: ``'cpu'``

    Returns:
    -------
        ``torch.Generator``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.utils.seed import get_torch_generator
        >>> generator = get_torch_generator(42)
        >>> torch.rand(2, 4, generator=generator)
        tensor([[0.8823, 0.9150, 0.3829, 0.9593],
                [0.3904, 0.6009, 0.2566, 0.7936]])
        >>> generator = get_torch_generator(42)
        >>> torch.rand(2, 4, generator=generator)
        tensor([[0.8823, 0.9150, 0.3829, 0.9593],
                [0.3904, 0.6009, 0.2566, 0.7936]])
    """
    generator = torch.Generator(device)
    generator.manual_seed(random_seed)
    return generator



from __future__ import annotations

__all__ = ["MultinomialSequenceGenerator", "UniformCategoricalSequenceGenerator"]

import math
from collections.abc import Sequence

import torch
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator
from startorch.utils.conversion import to_tuple
from startorch.utils.weight import prepare_probabilities


class MultinomialSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequences of categorical variables
    where each value is sampled from a multinomial distribution.

    Args:
    ----
        weights (``torch.Tensor`` of shape ``(num_categories,)`` and
            type float): Specifies the vector of weights associated
            at each category. The weights have to be positive but do
            not need to sum to 1.
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Multinomial
        >>> generator = Multinomial(weights=torch.ones(5))
        >>> generator
        MultinomialSequenceGenerator(num_categories=5, feature_size=(1,))
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        weights: torch.Tensor | Sequence[int | float],
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        self._probabilities = prepare_probabilities(weights)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(num_categories={self._probabilities.numel()}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            torch.multinomial(
                self._probabilities,
                batch_size * seq_len * math.prod(self._feature_size),
                replacement=True,
                generator=rng,
            ).view(batch_size, seq_len, *self._feature_size)
        )

    @classmethod
    def create_uniform_weights(cls, num_categories: int) -> MultinomialSequenceGenerator:
        r"""Initializes the weights with a uniform pattern.

        All the categories have the same probability.
        The weight of the ``i``-th category (``w_i``) is generated
        with the rule: ``w_i = 1``

        Args:
        ----
            num_categories (int): Specifies the number of categories.

        Returns:
        -------
            ``MultinomialSequenceGenerator``: A sequence generator where
                the weights of the multinomial distribution follow
                a uniform pattern.
        """
        return cls(weights=torch.ones(num_categories))

    @classmethod
    def create_linear_weights(cls, num_categories: int) -> MultinomialSequenceGenerator:
        r"""Initializes the weights with a linear pattern.

        The weight of the ``i``-th category (``w_i``) is generated
        with the rule: ``w_i = num_categories - i``

        Args:
        ----
            num_categories (int): Specifies the number of categories.

        Returns:
        -------
            ``MultinomialSequenceGenerator``: A sequence generator where
                the weights of the multinomial distribution follow a
                linear pattern.
        """
        return cls(
            weights=num_categories * torch.ones(num_categories) - torch.arange(num_categories)
        )

    @classmethod
    def create_exp_weights(
        cls, num_categories: int, scale: float = 0.1
    ) -> MultinomialSequenceGenerator:
        r"""Initializes the weights with an exponential pattern.

        The weight of the ``i``-th category (``w_i``) is generated with
        the rule: ``w_i = exp(-scale * i)``

        Args:
        ----
            num_categories (int): Specifies the number of categories.
            scale (float, optional): Specifies the scale parameter
                that controls the exponential function.
                Default: ``0.1``

        Returns:
        -------
            ``MultinomialSequenceGenerator``: A sequence generator where
                the weights of the multinomial distribution follow
                an exponential pattern.
        """
        return cls(weights=torch.arange(num_categories).float().mul(-scale).exp())


class UniformCategoricalSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequences of uniformly distributed
    categorical variables.

    All the categories have the same probability.

    Note: it is a more efficient implementation of
    ``Multinomial.generate_uniform_weights``.

    Args:
    ----
        num_categories (int): Specifies the number of categories.
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``tuple()``

    Raises:
    ------
        ValueError if ``num_categories`` is negative.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import UniformCategorical
        >>> generator = UniformCategorical(10)
        >>> generator
        UniformCategoricalSequenceGenerator(num_categories=10, feature_size=())
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        num_categories: int,
        feature_size: tuple[int, ...] | list[int] | int = tuple(),
    ) -> None:
        super().__init__()
        if num_categories <= 0:
            raise ValueError(
                f"num_categories has to be greater than 0 (received: {num_categories})"
            )
        self._num_categories = int(num_categories)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(num_categories={self._num_categories}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            torch.randint(
                low=0,
                high=self._num_categories,
                size=(batch_size, seq_len) + self._feature_size,
                generator=rng,
            )
        )



from __future__ import annotations

__all__ = [
    "NormalTensorGenerator",
    "RandNormalTensorGenerator",
    "RandTruncNormalTensorGenerator",
    "TruncNormalTensorGenerator",
]


from coola.utils.format import str_indent, str_mapping
from torch import Generator, Tensor

from startorch.random import normal, rand_normal, rand_trunc_normal, trunc_normal
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class NormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    Normal distribution.

    Args:
    ----
        mean (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the mean.
        std (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the standard
            deviation.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Normal, RandUniform
        >>> generator = Normal(
        ...     mean=RandUniform(low=-1.0, high=1.0), std=RandUniform(low=1.0, high=2.0)
        ... )
        >>> generator
        NormalTensorGenerator(
          (mean): RandUniformTensorGenerator(low=-1.0, high=1.0)
          (std): RandUniformTensorGenerator(low=1.0, high=2.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, mean: BaseTensorGenerator | dict, std: BaseTensorGenerator | dict) -> None:
        super().__init__()
        self._mean = setup_tensor_generator(mean)
        self._std = setup_tensor_generator(std)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"mean": self._mean, "std": self._std}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return normal(
            mean=self._mean.generate(size=size, rng=rng),
            std=self._std.generate(size=size, rng=rng),
            generator=rng,
        )


class RandNormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a sequence generator to generate cyclic sequences by
    sampling values from a Normal distribution.

    Args:
    ----
        mean (float, optional): Specifies the mean of the Normal
            distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of the
            Normal distribution. Default: ``1.0``

    Raises:
    ------
        ValueError if ``std`` is not a postive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandNormal
        >>> generator = RandNormal(mean=0.0, std=1.0)
        >>> generator
        RandNormalTensorGenerator(mean=0.0, std=1.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, mean: float = 0.0, std: float = 1.0) -> None:
        super().__init__()
        self._mean = float(mean)
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(mean={self._mean}, std={self._std})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_normal(
            size=size,
            mean=self._mean,
            std=self._std,
            generator=rng,
        )


class RandTruncNormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a sequence generator to generate cyclic sequences by
    sampling values from a truncated Normal distribution.

    Args:
    ----
        mean (float, optional): Specifies the mean of the Normal
            distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of
            the Normal distribution. Default: ``1.0``
        min_value (float, optional): Specifies the minimum value.
            Default: ``-3.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``3.0``

    Raises:
    ------
        ValueError if ``std`` is not a postive number.
        ValueError if ``max_value`` is lower than ``min_value``.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandTruncNormal
        >>> generator = RandTruncNormal(mean=0.0, std=1.0, min_value=-1.0, max_value=1.0)
        >>> generator
        RandTruncNormalTensorGenerator(mean=0.0, std=1.0, min_value=-1.0, max_value=1.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        mean: float = 0.0,
        std: float = 1.0,
        min_value: float = -3.0,
        max_value: float = 3.0,
    ) -> None:
        super().__init__()
        self._mean = float(mean)
        if std <= 0:
            raise ValueError(f"std has to be greater than 0 (received: {std})")
        self._std = float(std)
        if max_value < min_value:
            raise ValueError(
                f"max_value ({max_value}) has to be greater or equal to min_value ({min_value})"
            )
        self._min_value = float(min_value)
        self._max_value = float(max_value)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(mean={self._mean}, std={self._std}, "
            f"min_value={self._min_value}, max_value={self._max_value})"
        )

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_trunc_normal(
            size=size,
            mean=self._mean,
            std=self._std,
            min_value=self._min_value,
            max_value=self._max_value,
            generator=rng,
        )


class TruncNormalTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate sequence by sampling values from a
    truncated Normal distribution.

    Args:
    ----
        mean (``BaseTensorGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the mean.
        std (``BaseTensorGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the standard
            deviation.
        min_value (``BaseTensorGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            minimum value (included).
        max_value (``BaseTensorGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, TruncNormal
        >>> generator = TruncNormal(
        ...     mean=RandUniform(low=-1.0, high=1.0),
        ...     std=RandUniform(low=1.0, high=2.0),
        ...     min_value=RandUniform(low=-10.0, high=-5.0),
        ...     max_value=RandUniform(low=5.0, high=10.0),
        ... )
        >>> generator
        TruncNormalTensorGenerator(
          (mean): RandUniformTensorGenerator(low=-1.0, high=1.0)
          (std): RandUniformTensorGenerator(low=1.0, high=2.0)
          (min_value): RandUniformTensorGenerator(low=-10.0, high=-5.0)
          (max_value): RandUniformTensorGenerator(low=5.0, high=10.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        mean: BaseTensorGenerator | dict,
        std: BaseTensorGenerator | dict,
        min_value: BaseTensorGenerator | dict,
        max_value: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._mean = setup_tensor_generator(mean)
        self._std = setup_tensor_generator(std)
        self._min_value = setup_tensor_generator(min_value)
        self._max_value = setup_tensor_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "mean": self._mean,
                    "std": self._std,
                    "min_value": self._min_value,
                    "max_value": self._max_value,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return trunc_normal(
            mean=self._mean.generate(size=size, rng=rng),
            std=self._std.generate(size=size, rng=rng),
            min_value=self._min_value.generate(size=size, rng=rng),
            max_value=self._max_value.generate(size=size, rng=rng),
            generator=rng,
        )



from __future__ import annotations

__all__ = [
    "AbsSequenceGenerator",
    "AddScalarSequenceGenerator",
    "AddSequenceGenerator",
    "ClampSequenceGenerator",
    "CumsumSequenceGenerator",
    "DivSequenceGenerator",
    "ExpSequenceGenerator",
    "FmodSequenceGenerator",
    "LogSequenceGenerator",
    "MulScalarSequenceGenerator",
    "MulSequenceGenerator",
    "NegSequenceGenerator",
    "SqrtSequenceGenerator",
    "SubSequenceGenerator",
]

from collections.abc import Sequence

from coola.utils.format import str_indent, str_mapping, str_sequence
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.sequence.wrapper import BaseWrapperSequenceGenerator


class AbsSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the absolute value
    of a generated sequence.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Abs, RandNormal
        >>> generator = Abs(RandNormal())
        >>> generator
        AbsSequenceGenerator(
          (sequence): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).abs()


class AddSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator that adds multiple sequences.

    ``sequence = sequence_1 + sequence_2 + ... + sequence_n``

    Args:
    ----
        sequences (``Sequence``): Specifies the sequence generators or
            their configuration.

    Raises:
    ------
        ValueError if no sequence generator is provided.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Add, RandUniform, RandNormal
        >>> generator = Add([RandUniform(), RandNormal()])
        >>> generator
        AddSequenceGenerator(
          (0): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (1): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        sequences: Sequence[BaseSequenceGenerator | dict],
    ) -> None:
        super().__init__()
        if not sequences:
            raise ValueError(
                "No sequence generator. You need to specify at least one sequence generator"
            )
        self._sequences = tuple(setup_sequence_generator(generator) for generator in sequences)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_sequence(self._sequences))}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        output = self._sequences[0].generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        for sequence in self._sequences[1:]:
            output.add_(sequence.generate(seq_len=seq_len, batch_size=batch_size, rng=rng))
        return output


class AddScalarSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that adds a scalar value to a
    generated batch of sequences.

    Args:
    ----
        sequence (``BaseSequenceGenerator`` or dict):
            Specifies the sequence generator or its configuration.
        value (int or float): Specifies the scalar value to add.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import AddScalar, RandUniform, RandNormal
        >>> generator = AddScalar(RandUniform(), 42.0)
        >>> generator
        AddScalarSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (value): 42.0
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        generator: BaseSequenceGenerator | dict,
        value: int | float,
    ) -> None:
        super().__init__(generator=generator)
        self._value = value

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"sequence": self._generator, "value": self._value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        sequence = self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        sequence.add_(self._value)
        return sequence


class ClampSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator to generate a batch of sequences
    where the values are clamped.

    Note: ``min_value`` and ``max_value`` cannot be both ``None``.

    Args:
    ----
        sequence (``BaseSequenceGenerator`` or dict):
            Specifies the sequence generator or its configuration.
        min (int, float or ``None``): Specifies the lower bound.
            If ``min_value`` is ``None``, there is no lower bound.
        max (int, float or ``None``): Specifies the upper bound.
            If ``max_value`` is ``None``, there is no upper bound.

    Raises:
    ------
        ValueError if both ``min`` and ``max`` are ``None``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Clamp, RandNormal
        >>> generator = Clamp(RandNormal(), -1.0, 1.0)
        >>> generator
        ClampSequenceGenerator(
          (sequence): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
          (min): -1.0
          (max): 1.0
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        generator: BaseSequenceGenerator | dict,
        min: int | float | None,  # noqa: A002
        max: int | float | None,  # noqa: A002
    ) -> None:
        super().__init__(generator=generator)
        if min is None and max is None:
            raise ValueError("`min` and `max` cannot be both None")
        self._min = min
        self._max = max

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"sequence": self._generator, "min": self._min, "max": self._max})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).clamp(
            min=self._min, max=self._max
        )


class CumsumSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the cumulative sum
    of a generated sequence.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Cumsum, RandUniform
        >>> generator = Cumsum(RandUniform())
        >>> generator
        CumsumSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(
            seq_len=seq_len, batch_size=batch_size, rng=rng
        ).cumsum_along_seq()


class DivSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator that divides one sequence by
    another one.

    ``sequence = dividend / divisor`` (a.k.a. true division)

    Args:
    ----
        dividend (``Sequence``): (``BaseSequenceGenerator`` or dict):
            Specifies the dividend sequence generator or its
            configuration.
        divisor (``Sequence``): (``BaseSequenceGenerator`` or dict):
            Specifies the divisor sequence generator or its
            configuration.
        rounding_mode (str or ``None``, optional): Specifies the
            type of rounding applied to the result.
            - ``None``: true division.
            - ``"trunc"``: rounds the results of the division
                towards zero.
            - ``"floor"``: floor division.
            Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Div, RandUniform, RandNormal
        >>> generator = Div(RandNormal(), RandUniform(1.0, 10.0))
        >>> generator
        DivSequenceGenerator(
          (dividend): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
          (divisor): RandUniformSequenceGenerator(low=1.0, high=10.0, feature_size=(1,))
          (rounding_mode): None
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        dividend: BaseSequenceGenerator | dict,
        divisor: BaseSequenceGenerator | dict,
        rounding_mode: str | None = None,
    ) -> None:
        super().__init__()
        self._dividend = setup_sequence_generator(dividend)
        self._divisor = setup_sequence_generator(divisor)
        self._rounding_mode = rounding_mode

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "dividend": self._dividend,
                    "divisor": self._divisor,
                    "rounding_mode": self._rounding_mode,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._dividend.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).div(
            self._divisor.generate(seq_len=seq_len, batch_size=batch_size, rng=rng),
            rounding_mode=self._rounding_mode,
        )


class ExpSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the exponential of
    a batch of sequences.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Exp, RandUniform, RandNormal
        >>> generator = Exp(RandUniform())
        >>> generator
        ExpSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).exp()


class FmodSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a tensor generator that computes the element-wise
    remainder of division.

    Args:
    ----
        dividend (``BaseSequenceGenerator`` or dict):
            Specifies the sequence generator (or its configuration) that
            generates the dividend values.
        divisor (int or float): Specifies the divisor.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Fmod, RandUniform
        >>> generator = Fmod(dividend=RandUniform(low=-100, high=100), divisor=10.0)
        >>> generator
        FmodSequenceGenerator(
          (dividend): RandUniformSequenceGenerator(low=-100.0, high=100.0, feature_size=(1,))
          (divisor): 10.0
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
        >>> generator = Fmod(
        ...     dividend=RandUniform(low=-100, high=100), divisor=RandUniform(low=1, high=10)
        ... )
        >>> generator
        FmodSequenceGenerator(
          (dividend): RandUniformSequenceGenerator(low=-100.0, high=100.0, feature_size=(1,))
          (divisor): RandUniformSequenceGenerator(low=1.0, high=10.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        dividend: BaseSequenceGenerator | dict,
        divisor: BaseSequenceGenerator | dict | int | float,
    ) -> None:
        super().__init__()
        self._dividend = setup_sequence_generator(dividend)
        self._divisor = setup_sequence_generator(divisor) if isinstance(divisor, dict) else divisor

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"dividend": self._dividend, "divisor": self._divisor}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        seq = self._dividend.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        divisor = self._divisor
        if isinstance(divisor, BaseSequenceGenerator):
            divisor = divisor.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        seq.fmod_(divisor)
        return seq


class LogSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the logarithm of a
    batch of sequences.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Log, RandUniform, RandNormal
        >>> generator = Log(RandUniform(1.0, 10.0))
        >>> generator
        LogSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=1.0, high=10.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).log()


class MulSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator that multiplies multiple
    sequences.

    ``sequence = sequence_1 * sequence_2 * ... * sequence_n``

    Args:
    ----
        sequences (``Sequence``): Specifies the sequence generators.

    Raises:
    ------
        ValueError if no sequence generator is provided.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Mul, RandUniform, RandNormal
        >>> generator = Mul([RandUniform(), RandNormal()])
        >>> generator
        MulSequenceGenerator(
          (0): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (1): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        sequences: Sequence[BaseSequenceGenerator | dict],
    ) -> None:
        super().__init__()
        if not sequences:
            raise ValueError(
                "No sequence generator. You need to specify at least one sequence generator"
            )
        self._sequences = tuple(setup_sequence_generator(generator) for generator in sequences)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_sequence(self._sequences))}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        output = self._sequences[0].generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        for generator in self._sequences[1:]:
            output.mul_(generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng))
        return output


class MulScalarSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that multiplies a scalar value to
    a generated batch of sequences.

    Args:
    ----
        sequence (``BaseSequenceGenerator`` or dict):
            Specifies the sequence generator or its configuration.
        value (int or float): Specifies the scalar value to multiply.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import MulScalar, RandUniform, RandNormal
        >>> generator = MulScalar(RandUniform(), 2.0)
        >>> generator
        MulScalarSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (value): 2.0
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        generator: BaseSequenceGenerator | dict,
        value: int | float,
    ) -> None:
        super().__init__(generator=generator)
        self._value = value

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"sequence": self._generator, "value": self._value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        sequence = self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        sequence.mul_(self._value)
        return sequence


class NegSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the negation of a
    generated sequence.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Neg, RandUniform, RandNormal
        >>> generator = Neg(RandUniform())
        >>> generator
        NegSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return -self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)


class SqrtSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the squared root of
    a batch of sequences.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Sqrt, RandUniform, RandNormal
        >>> generator = Sqrt(RandUniform(1.0, 4.0))
        >>> generator
        SqrtSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=1.0, high=4.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).sqrt()


class SubSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator that subtracts sequences.

    ``sequence = sequence_1 - sequence_2``

    Args:
    ----
        sequence1 (``BaseSequenceGenerator`` or dict):
            Specifies the first sequence generator or its
            configuration.
        sequence2 (``BaseSequenceGenerator`` or dict):
            Specifies the second sequence generator or its
            configuration.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Sub, RandUniform, RandNormal
        >>> generator = Sub(RandUniform(), RandNormal())
        >>> generator
        SubSequenceGenerator(
          (sequence1): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (sequence2): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        sequence1: BaseSequenceGenerator | dict,
        sequence2: BaseSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._sequence1 = setup_sequence_generator(sequence1)
        self._sequence2 = setup_sequence_generator(sequence2)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"sequence1": self._sequence1, "sequence2": self._sequence2}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._sequence1.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).sub(
            self._sequence2.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        )
